import pytest

import torch

from syft.frameworks.torch.tensors.interpreters.large_precision import LargePrecisionTensor


def test_wrap(workers):
    """
    Test the .on() wrap functionality for LargePrecisionTensor
    """
    x_tensor = torch.Tensor([1, 2, 3])
    x = LargePrecisionTensor().on(x_tensor)
    assert isinstance(x, torch.Tensor)
    assert isinstance(x.child, LargePrecisionTensor)
    assert isinstance(x.child.child, torch.Tensor)


def test_fix_prec(workers):
    x = torch.tensor([1.5, 2.0, 3.0])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_2d_tensors(workers):
    x = torch.tensor([[1.5, 2.0, 3.0], [4.5, 5.0, 6.0]])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_3d_tensors(workers):
    x = torch.tensor([[[1.5, 2.0, 3.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_negative_numbers(workers):
    x = torch.tensor([[[-1.5, 2.0, 3.0]], [[4.5, 5.0, 6.0]], [[7.0, 8.0, 9.0]]])
    enlarged = x.fix_prec(
        base=10, internal_type=torch.int16, precision_fractional=128, verbose=True
    )
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_add_multiple_dimensions(workers):
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.5, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.5, 5.0, 6.0]]])
    lpt1 = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    lpt2 = y.fix_prec(internal_type=torch.int16, precision_fractional=128)
    expected = torch.tensor([[[-3.0, -4.0, -6.0]], [[9.0, 10.0, 3.0]]])
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_add_negative_values():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([-90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_add():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([30.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_iadd():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([30.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.add_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_add_different_dims():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([100000.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([100020.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 + lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_mul():
    internal_type = torch.int16
    precision_fractional = 32
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([200.0])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 * lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_imul():
    internal_type = torch.int16
    precision_fractional = 32
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([20.0])
    expected = torch.tensor([200.0])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.mul_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_mul_multiple_dims():
    internal_type = torch.int16
    precision_fractional = 32
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, 6.0]]])
    expected = torch.tensor([[[2.25, 4.0, 9.0]], [[16.0, 25.0, -18.0]]])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1 = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = y.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 * lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_concat_ops():
    internal_type = torch.int16
    precision_fractional = 32
    x = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, -3.0]]])
    y = torch.tensor([[[-1.5, -2.0, -3.0]], [[4.0, 5.0, 6.0]]])
    z = torch.tensor([[[-1.0, -2.0, 2.5]], [[4.0, -5.0, 7.5]]])
    expected = torch.tensor([[[1.25, 2.0, 11.5]], [[20.0, 20.0, -10.5]]])
    expected.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_x = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_y = y.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt_z = z.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = (lpt_x * lpt_y) + lpt_z

    assert torch.all(torch.eq(expected, result.float_precision()))


def test_uint8_representation(workers):
    x = torch.tensor([[1.5, 2.0, 3.0], [4.5, 5.0, 6.0]])
    enlarged = x.fix_prec(internal_type=torch.uint8, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


def test_sub():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 - lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


def test_isub():
    internal_type = torch.int16
    precision_fractional = 128
    x1 = torch.tensor([10.0])
    x2 = torch.tensor([90000000000000010.0])
    expected = torch.tensor([-90000000000000000.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt1.sub_(lpt2)
    assert torch.all(torch.eq(expected, lpt1.float_precision()))


def test_diff_dims_in_same_tensor():
    internal_type = torch.int16
    precision_fractional = 128
    x = torch.tensor([2000.0, 1.0])
    lpt_x = x.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    restored = lpt_x.float_precision()
    assert torch.all(torch.eq(x, restored))


def test_mod():
    internal_type = torch.int16
    precision_fractional = 128
    expected = torch.tensor([6.0, 3.0])
    x1 = torch.tensor([6.0, 12.0])
    x2 = torch.tensor([9.0])
    lpt1 = x1.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    lpt2 = x2.fix_prec(internal_type=internal_type, precision_fractional=precision_fractional)
    result = lpt1 % lpt2
    assert torch.all(torch.eq(expected, result.float_precision()))


test_data = [
    (torch.tensor([1]), torch.tensor([1.0])),
    (torch.tensor([1.0]), torch.tensor([1.0])),
    (torch.tensor([2000.0, 1.0]), torch.tensor([2000.0, 1.0])),
    (torch.tensor([2000.0, 1]), torch.tensor([2000.0, 1.0])),
    (torch.tensor([-2000.0]), torch.tensor([-2000.0])),
    (torch.tensor([-2000.0, -50]), torch.tensor([-2000.0, -50.0])),
    (torch.tensor([-2000.0, 50]), torch.tensor([-2000.0, 50.0])),
    (
        torch.tensor([[-2000.0, 50], [1000.5, -25]]),
        torch.tensor([[-2000.0, 50.0], [1000.5, -25.0]]),
    ),
    (torch.tensor([-2000.0123458910]), torch.tensor([-2000.0123458910])),
    (torch.tensor([2000.0123458910]), torch.tensor([2000.0123458910])),
]


@pytest.mark.parametrize("x, expected", test_data)
def test_types(x, expected):
    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(expected, restored))


@pytest.mark.parametrize("x, expected", test_data)
def test_share_and_get(workers, x, expected):
    alice, bob, james = (workers["alice"], workers["bob"], workers["james"])

    enlarged = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    expected = expected.fix_prec(internal_type=torch.int16, precision_fractional=128)

    shared = enlarged.share(alice, bob, crypto_provider=james)

    result = shared.get().float_precision()

    assert (expected.float_precision() == result).all()


def test_share_add(workers):
    alice, bob, james = (workers["alice"], workers["bob"], workers["james"])

    x = torch.tensor([5.0])
    expected = x + x

    x = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    x_shared = x.share(alice, bob, crypto_provider=james)

    y = (x_shared + x_shared).get().float_precision()

    assert torch.all(torch.eq(expected, y))


def test_share_sub(workers):
    alice, bob, james = (workers["alice"], workers["bob"], workers["james"])

    x = torch.tensor([5.0])
    expected = x - x

    x = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
    x_shared = x.share(alice, bob, crypto_provider=james)

    y = (x_shared - x_shared).get().float_precision()

    assert torch.all(torch.eq(expected, y))


def test_storage():
    x = torch.tensor([1.0, 2.0, 3.0])
    enlarged = x.fix_prec(storage="large")
    restored = enlarged.float_precision()
    # And now x and restored must be the same
    assert torch.all(torch.eq(x, restored))


# def test_share_mul(workers):
#     alice, bob, james = (workers["alice"], workers["bob"], workers["james"])
#
#     x = torch.tensor([5.0])
#     expected = x * x
#
#     x = x.fix_prec(internal_type=torch.int16, precision_fractional=128)
#     x_shared = x.share(alice, bob, crypto_provider=james)
#
#     # TODO
#     #  This is tricky. The multiplication of the shares in the workers will not yield the expected results
#     #  because they have the original number decomposed into parts.
#     y = (x_shared * x_shared).get().float_precision()
#
#     assert torch.all(torch.eq(expected, y))



import torch
import syft


def test_init():
    hook = syft.TorchHook(torch, verbose=True)
    tensor_extension = torch.Tensor()
    assert tensor_extension.id is not None
    assert tensor_extension.owner is not None



import tf_encrypted as tfe

from syft.frameworks.keras.model import serve
from syft.frameworks.keras.model import share
from syft.frameworks.keras.model import stop
from syft.frameworks.keras.layers import add_constructor_registration
from syft.frameworks.keras.layers import filter_layers


class KerasHook:
    def __init__(self, keras):
        self.keras = keras
        if not hasattr(keras, "_hooked"):
            self._hook_layers()
            self._hook_sequential()
            self.keras._hooked = True

    def _hook_layers(self):
        for layer_cls in filter_layers(self.keras.layers, tfe.keras.layers):
            layer_cls = add_constructor_registration(layer_cls)

    def _hook_sequential(self):
        seq_cls = getattr(self.keras, "Sequential")
        setattr(seq_cls, "share", share)
        setattr(seq_cls, "serve", serve)
        setattr(seq_cls, "stop", stop)



import numpy as np
import torch
from torch import nn
from torch.nn import init

import syft
from syft.frameworks.torch.tensors.interpreters.precision import FixedPrecisionTensor
from syft.frameworks.torch.tensors.interpreters.additive_shared import AdditiveSharingTensor
from syft.generic.pointers.pointer_tensor import PointerTensor


class RNNCellBase(nn.Module):
    """
    Cell to be used as base for all RNN cells, including GRU and LSTM
    This class overrides the torch.nn.RNNCellBase
    Only Linear and Dropout layers are used to be able to use MPC
    """

    def __init__(self, input_size, hidden_size, bias, num_chunks, nonlinearity=None):
        super(RNNCellBase, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.num_chunks = num_chunks
        self.nonlinearity = nonlinearity
        self.fc_xh = nn.Linear(input_size, self.num_chunks * hidden_size, bias=bias)
        self.fc_hh = nn.Linear(hidden_size, self.num_chunks * hidden_size, bias=bias)
        self.reset_parameters()

    def reset_parameters(self):
        """
        This method initializes or reset all the parameters of the cell.
        The paramaters are initiated following a uniform distribution.
        """
        std = 1.0 / np.sqrt(self.hidden_size)
        for w in self.parameters():
            init.uniform_(w, -std, std)

    def init_hidden(self, input):
        """
        This method initializes a hidden state when no hidden state is provided
        in the forward method. It creates a hidden state with zero values.
        """
        h = torch.zeros(input.shape[0], self.hidden_size, dtype=input.dtype, device=input.device)
        if input.has_child() and isinstance(input.child, PointerTensor):
            h = h.send(input.child.location)
        if input.has_child() and isinstance(input.child, FixedPrecisionTensor):
            h = h.fix_precision()
            child = input.child
            if isinstance(child.child, AdditiveSharingTensor):
                crypto_provider = child.child.crypto_provider
                owners = child.child.locations
                h = h.share(*owners, crypto_provider=crypto_provider)
        return h


class RNNCell(RNNCellBase):
    """
    Python implementation of RNNCell with tanh or relu non-linearity for MPC
    This class overrides the torch.nn.RNNCell
    """

    def __init__(self, input_size, hidden_size, bias=True, nonlinearity="tanh"):
        super(RNNCell, self).__init__(input_size, hidden_size, bias, num_chunks=1)

        if nonlinearity == "tanh":
            self.nonlinearity = torch.tanh
        elif nonlinearity == "relu":
            self.nonlinearity = torch.relu
        else:
            raise ValueError("Unknown nonlinearity: {}".format(nonlinearity))

    def forward(self, x, h=None):

        if h is None:
            h = self.init_hidden(x)
        h_ = self.nonlinearity(self.fc_xh(x) + self.fc_hh(h))

        return h_


class GRUCell(RNNCellBase):
    """
    Python implementation of GRUCell for MPC
    This class overrides the torch.nn.GRUCell
    """

    def __init__(self, input_size, hidden_size, bias=True, nonlinearity=None):
        super(GRUCell, self).__init__(input_size, hidden_size, bias, num_chunks=3)

    def forward(self, x, h=None):

        if h is None:
            h = self.init_hidden(x)

        gate_x = self.fc_xh(x)
        gate_h = self.fc_hh(h)
        x_r, x_z, x_n = gate_x.chunk(self.num_chunks, 1)
        h_r, h_z, h_n = gate_h.chunk(self.num_chunks, 1)

        resetgate = torch.sigmoid(x_r + h_r)
        updategate = torch.sigmoid(x_z + h_z)
        newgate = torch.tanh(x_n + (resetgate * h_n))

        h_ = newgate + updategate * (h - newgate)

        return h_


class LSTMCell(RNNCellBase):
    """
    Python implementation of LSTMCell for MPC
    This class overrides the torch.nn.LSTMCell
    """

    def __init__(self, input_size, hidden_size, bias=True, nonlinearity=None):
        super(LSTMCell, self).__init__(input_size, hidden_size, bias, num_chunks=4)

    def reset_parameters(self):
        super(LSTMCell, self).reset_parameters()

        # Bias of forget gate should be initialize with 1 or 2
        # Ref: http://proceedings.mlr.press/v37/jozefowicz15.pdf
        incr_bias = 1.0 / self.hidden_size
        init.constant_(self.fc_xh.bias[self.hidden_size : 2 * self.hidden_size], incr_bias)
        init.constant_(self.fc_hh.bias[self.hidden_size : 2 * self.hidden_size], incr_bias)

    def forward(self, x, hc=None):

        if hc is None:
            hc = (self.init_hidden(x), self.init_hidden(x))
        h, c = hc

        gate_x = self.fc_xh(x)
        gate_h = self.fc_hh(h)

        x_i, x_f, x_c, x_o = gate_x.chunk(self.num_chunks, 1)
        h_i, h_f, h_c, h_o = gate_h.chunk(self.num_chunks, 1)

        inputgate = torch.sigmoid(x_i + h_i)
        forgetgate = torch.sigmoid(x_f + h_f)
        cellgate = torch.tanh(x_c + h_c)
        outputgate = torch.sigmoid(x_o + h_o)

        c_ = torch.mul(forgetgate, c) + torch.mul(inputgate, cellgate)

        h_ = torch.mul(outputgate, torch.tanh(c_))

        return h_, c_


class RNNBase(nn.Module):
    """
    Module to be used as base for all RNN modules, including GRU and LSTM
    This class overrides the torch.nn.RNNBase
    Only Linear and Dropout layers are used to be able to use MPC
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers,
        bias,
        batch_first,
        dropout,
        bidirectional,
        base_cell,
        nonlinearity=None,
    ):
        super(RNNBase, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = float(dropout)
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1
        self.is_lstm = base_cell is LSTMCell
        self.nonlinearity = nonlinearity

        # Dropout layers
        # TODO: implement a nn.Dropout class for PySyft
        # Link to issue: https://github.com/OpenMined/PySyft/issues/2500

        # Build RNN layers
        self.rnn_forward = nn.ModuleList()
        for layer in range(self.num_layers):
            if layer == 0:
                self.rnn_forward.append(base_cell(input_size, hidden_size, bias, nonlinearity))
            else:
                self.rnn_forward.append(base_cell(hidden_size, hidden_size, bias, nonlinearity))

        if self.bidirectional:
            self.rnn_backward = nn.ModuleList()
            for layer in range(self.num_layers):
                if layer == 0:
                    self.rnn_backward.append(base_cell(input_size, hidden_size, bias, nonlinearity))
                else:
                    self.rnn_backward.append(
                        base_cell(hidden_size, hidden_size, bias, nonlinearity)
                    )

    def forward(self, x, h=None):
        # If batch_first == True, swap axis with seq_len
        # At the end of the process we swap it back to the original structure
        if self.batch_first:
            x, h = self._swap_axis(x, h)

        # If it is LSTM, get hidden and cell states
        if h is not None and self.is_lstm:
            h, c = h

        batch_size = x.shape[1]
        seq_len = x.shape[0]

        # Initiate states if needed
        if h is None:
            h = self._init_hidden(x)
            c = self._init_hidden(x) if self.is_lstm else None

        # If bidirectional==True, split states in two, each one for each direction
        if self.bidirectional:
            h = h.contiguous().view(self.num_layers, 2, batch_size, self.hidden_size)
            h_for = h[:, 0, :, :]
            h_back = h[:, 1, :, :]
            if self.is_lstm:
                c = c.contiguous().view(self.num_layers, 2, batch_size, self.hidden_size)
                c_for = c[:, 0, :, :]
                c_back = c[:, 1, :, :]
            else:
                c_for = c_back = None

        else:
            h_for = h
            c_for = c if self.is_lstm else None

        # Run through rnn in the forward direction
        output = x.new(seq_len, batch_size, self.hidden_size).zero_()
        for t in range(seq_len):
            h_for, c_for = self._apply_time_step(x, h_for, c_for, t)
            output[t, :, :] = h_for[-1, :, :]

        hidden = h_for
        cell = c_for  # None if it is not an LSTM

        # Run through rnn in the backward direction if bidirectional==True
        if self.bidirectional:
            output_back = x.new(seq_len, batch_size, self.hidden_size).zero_()
            for t in range(seq_len - 1, -1, -1):
                h_back, c_back = self._apply_time_step(x, h_back, c_back, t, reverse_direction=True)
                output_back[t, :, :] = h_back[-1, :, :]

            # Concatenate both directions
            output = torch.cat((output, output_back), dim=-1)
            hidden = torch.cat((hidden, h_back), dim=0)
            if self.is_lstm:
                cell = torch.cat((cell, c_back), dim=0)

        # If batch_first == True, swap axis back to get original structure
        if self.batch_first:
            output = torch.transpose(output, 0, 1)
            hidden = torch.transpose(hidden, 0, 1)
            if self.is_lstm:
                cell = torch.transpose(cell, 0, 1)

        hidden = (hidden, cell) if self.is_lstm else hidden

        return output, hidden

    def _swap_axis(self, x, h):
        """
        This method swap the axes for batch_size and seq_len. It is used when
        batch_first==True.
        """
        x = torch.transpose(x, 0, 1)
        if h is not None:
            if self.is_lstm:
                h, c = h
                h = torch.transpose(h, 0, 1)
                c = torch.transpose(c, 0, 1)
                h = (h, c)
            else:
                h = torch.transpose(h, 0, 1)
        return x, h

    def _init_hidden(self, input):
        """
        This method initializes a hidden state when no hidden state is provided
        in the forward method. It creates a hidden state with zero values for each
        layer of the network.
        """
        h = torch.zeros(
            self.num_layers * self.num_directions,
            input.shape[1],
            self.hidden_size,
            dtype=input.dtype,
            device=input.device,
        )
        if input.has_child() and isinstance(input.child, PointerTensor):
            h = h.send(input.child.location)
        if input.has_child() and isinstance(input.child, FixedPrecisionTensor):
            h = h.fix_precision()
            child = input.child
            if isinstance(child.child, AdditiveSharingTensor):
                crypto_provider = child.child.crypto_provider
                owners = child.child.locations
                h = h.share(*owners, crypto_provider=crypto_provider)
        return h

    def _apply_time_step(self, x, h, c, t, reverse_direction=False):
        """
        Apply RNN layers at time t, given input and previous hidden states
        """
        rnn_layers = self.rnn_backward if reverse_direction else self.rnn_forward

        h_next = h.new(h.shape).zero_()
        c_next = c.new(c.shape).zero_() if self.is_lstm else None

        for layer in range(self.num_layers):
            if layer == 0:
                if self.is_lstm:
                    h_next[layer, :, :], c_next[layer, :, :] = rnn_layers[layer](
                        x[t, :, :], (h[layer, :, :], c[layer, :, :])
                    )
                else:
                    h_next[layer, :, :] = rnn_layers[layer](x[t, :, :], h[layer, :, :])
            else:
                if self.is_lstm:
                    h_next[layer, :, :], c_next[layer, :, :] = rnn_layers[layer](
                        h_next[layer - 1, :, :].clone(), (h[layer, :, :], c[layer, :, :])
                    )
                else:
                    h_next[layer, :, :] = rnn_layers[layer](
                        h_next[layer - 1, :, :].clone(), h[layer, :, :]
                    )

        return h_next, c_next


class RNN(RNNBase):
    """
    Python implementation of RNN for MPC
    This class overrides the torch.nn.RNN
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=1,
        nonlinearity="tanh",
        bias=True,
        batch_first=False,
        dropout=0,
        bidirectional=False,
    ):

        super(RNN, self).__init__(
            input_size,
            hidden_size,
            num_layers,
            bias,
            batch_first,
            dropout,
            bidirectional,
            RNNCell,
            nonlinearity,
        )


class GRU(RNNBase):
    """
    Python implementation of GRU for MPC
    This class overrides the torch.nn.GRU
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=1,
        bias=True,
        batch_first=False,
        dropout=0,
        bidirectional=False,
    ):

        super(GRU, self).__init__(
            input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional, GRUCell
        )


class LSTM(RNNBase):
    """
    Python implementation of LSTM for MPC
    This class overrides the torch.nn.LSTM
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=1,
        bias=True,
        batch_first=False,
        dropout=0,
        bidirectional=False,
    ):

        super(LSTM, self).__init__(
            input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional, LSTMCell
        )



import sys
import pytest
from syft import dependency_check


@pytest.mark.skipif(not dependency_check.tensorflow_available, reason="tf 2.0+ not installed")
def test_tensorflow_available():  # pragma: no cover
    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    from syft import dependency_check

    assert dependency_check.tensorflow_available


@pytest.mark.skipif(not dependency_check.tfe_available, reason="tf_encrypted not installed")
def test_tf_encrypted_available():  # pragma: no cover
    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    from syft import dependency_check

    assert dependency_check.tfe_available


@pytest.mark.skipif(not dependency_check.torch_available, reason="torch not installed")
def test_torch_available():  # pragma: no cover
    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    from syft import dependency_check

    assert dependency_check.torch_available


@pytest.mark.usefixtures("hide_module")
def test_tensorflow_missing():  # pragma: no cover
    sys.modules.pop("syft", None)
    sys.modules.pop("syft.dependency_check", None)
    sys.modules.pop("tensorflow", None)
    sys.modules.pop("tf", None)
    import syft.dependency_check

    assert not syft.dependency_check.tensorflow_available


@pytest.mark.usefixtures("hide_module")
def test_tf_encrypted_missing():  # pragma: no cover
    sys.modules.pop("syft.dependency_check", None)
    sys.modules.pop("tf_encrypted", None)
    sys.modules.pop("tfe", None)
    import syft.dependency_check

    assert not syft.dependency_check.tfe_available



import time
from functools import wraps


def assert_time(max_time):
    """
    Decorator used to assert time execution of functions.
    Args:
        max_time: int or float. Maximum time in seconds that the decorated
        function should take to be executed
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            t0 = time.time()
            func(*args, **kwargs)
            dt = time.time() - t0
            assert dt < max_time

        return wrapper

    return decorate



import torch
from syft.generic.pointers import callable_pointer
from syft.generic.pointers.object_wrapper import ObjectWrapper


def test_create_callable_pointer(workers):
    alice = workers["alice"]
    bob = workers["bob"]
    callable_pointer.create_callable_pointer(
        id=500,
        id_at_location=2,
        location=alice,
        owner=bob,
        tags="tags",
        description="description",
        register_pointer=True,
    )

    assert len(alice._objects) == 0
    assert len(bob._objects) == 1

    callable_pointer.create_callable_pointer(
        id=501,
        id_at_location=2,
        location=alice,
        owner=bob,
        tags="tags",
        description="description",
        register_pointer=False,
    )

    assert len(alice._objects) == 0
    assert len(bob._objects) == 1


def test_get_obj_callable_pointer(workers):
    alice = workers["alice"]
    bob = workers["bob"]

    x = torch.tensor(5)
    x_ptr = x.send(alice)

    obj_ptr = callable_pointer.create_callable_pointer(
        id=1,
        id_at_location=x_ptr.id_at_location,
        location=alice,
        owner=bob,
        tags="tags",
        description="description",
        register_pointer=True,
    )

    assert len(alice._objects) == 1
    assert len(bob._objects) == 1

    x_get = obj_ptr.get()

    assert len(alice._objects) == 0
    assert len(bob._objects) == 0
    assert x_get == x


def test_call_callable_pointer(workers):
    def foo(x):
        return x + 2

    alice = workers["alice"]
    bob = workers["bob"]

    id_alice = 100
    id_bob = 200
    foo_wrapper = ObjectWrapper(id=id_alice, obj=foo)

    alice.register_obj(foo_wrapper, id_alice)

    foo_ptr = callable_pointer.create_callable_pointer(
        id=id_bob,
        id_at_location=id_alice,
        location=alice,
        owner=bob,
        tags="tags",
        description="description",
        register_pointer=True,
    )

    res = foo_ptr(4)

    assert res == 6



import logging
from syft import dependency_check

logger = logging.getLogger(__name__)

__all__ = list()

if dependency_check.tensorflow_available:
    from syft.frameworks import tensorflow

    __all__.append("tensorflow")

if dependency_check.tfe_available:
    from syft.frameworks import keras

    __all__.append("keras")

if dependency_check.torch_available:
    from syft.frameworks import torch

    __all__.append("torch")



"""
This file exists as the Python encoding of all Message types that Syft sends over the network. It is
an important bottleneck in the system, impacting both security, performance, and cross-platform
compatability. As such, message types should strive to not be framework specific (i.e., Torch,
Tensorflow, etc.).

All Syft message types extend the Message class.
"""

import syft as sy
from syft import codes
from syft.workers.abstract import AbstractWorker


class Message:
    """All syft message types extend this class

    All messages in the pysyft protocol extend this class. This abstraction
    requires that every message has an integer type, which is important because
    this integer is what determines how the message is handled when a BaseWorker
    receives it.

    Additionally, this type supports a default simplifier and detailer, which are
    important parts of PySyft's serialization and deserialization functionality.
    You can read more abouty detailers and simplifiers in syft/serde/serde.py.
    """

    def __init__(self, msg_type: int, contents=None):
        self.msg_type = msg_type

        # saves us a write op but costs us a check op to only sometimes
        # set ._contents
        if contents is not None:
            self._contents = contents

    @property
    def contents(self):
        """Return a tuple with the contents of the message (backwards compatability)

        Some of our codebase still assumes that all message types have a .contents attribute. However,
        the contents attribute is very opaque in that it doesn't put any constraints on what the contents
        might be. Some message types can be more efficient by storing their contents more explicitly (see
        Operation). They can override this property to return a tuple view on their other properties.
        """
        if hasattr(self, "_contents"):
            return self._contents
        else:
            return None

    def _simplify(self):
        return (self.msg_type, self.contents)

    @staticmethod
    def simplify(ptr: "Message") -> tuple:
        """
        This function takes the attributes of a Message and saves them in a tuple.
        The detail() method runs the inverse of this method.
        Args:
            ptr (Message): a Message
        Returns:
            tuple: a tuple holding the unique attributes of the message
        Examples:
            data = simplify(ptr)
        """

        return (ptr.msg_type, sy.serde._simplify(ptr.contents))

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "Message":
        """
        This function takes the simplified tuple version of this message and converts
        it into a message. The simplify() method runs the inverse of this method.

        This method shouldn't get called very often. It exists as a backup but in theory
        every message type should have its own detailer.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (Message): a Operation.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """

        # TODO: attempt to use the msg_tuple[0] to return the correct type instead of Message
        # https://github.com/OpenMined/PySyft/issues/2514
        # TODO: as an alternative, this detailer could raise NotImplementedException
        # https://github.com/OpenMined/PySyft/issues/2514

        return Message(msg_tuple[0], sy.serde._detail(worker, msg_tuple[1]))

    def __str__(self):
        """Return a human readable version of this message"""
        return f"({codes.code2MSGTYPE[self.msg_type]} {self.contents})"

    def __repr__(self):
        """Return a human readable version of this message"""
        return self.__str__()


class Operation(Message):
    """All syft operations use this message type

    In Syft, an operation is when one worker wishes to tell another worker to do something with
    objects contained in the worker._objects registry (or whatever the official object store is
    backed with in the case that it's been overridden). Semantically, one could view all Messages
    as a kind of operation, but when we say operation this is what we mean. For example, telling a
    worker to take two tensors and add them together is an operation. However, sending an object
    from one worker to another is not an operation (and would instead use the ObjectMessage type)."""

    def __init__(self, message, return_ids):
        """Initialize an operation message

        Args:
            message (Tuple): this is typically the args and kwargs of a method call on the client, but it
                can be any information necessary to execute the operation properly.
            return_ids (Tuple): primarily for our async infrastructure (Plan, Protocol, etc.), the id of
                operation results are set by the client. This allows the client to be able to predict where
                the results will be ahead of time. Importantly, this allows the client to pre-initalize the
                pointers to the future data, regardless of whether the operation has yet executed. It also
                reduces the size of the response from the operation (which is very often empty).

        """

        # call the parent constructor - setting the type integer correctly
        super().__init__(codes.MSGTYPE.CMD)

        self.message = message
        self.return_ids = return_ids

    @property
    def contents(self):
        """Return a tuple with the contents of the operation (backwards compatability)

        Some of our codebase still assumes that all message types have a .contents attribute. However,
        the contents attribute is very opaque in that it doesn't put any constraints on what the contents
        might be. Since we know this message is a operation, we instead choose to store contents in two pieces,
        self.message and self.return_ids, which allows for more efficient simplification (we don't have to
        simplify return_ids because they are always a list of integers, meaning they're already simplified)."""

        return (self.message, self.return_ids)

    @staticmethod
    def simplify(ptr: "Operation") -> tuple:
        """
        This function takes the attributes of a Operation and saves them in a tuple
        Args:
            ptr (Operation): a Message
        Returns:
            tuple: a tuple holding the unique attributes of the message
        Examples:
            data = simplify(ptr)
        """
        # NOTE: we can skip calling _simplify on return_ids because they should already be
        # a list of simple types.
        return (ptr.msg_type, (sy.serde._simplify(ptr.message), ptr.return_ids))

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "Operation":
        """
        This function takes the simplified tuple version of this message and converts
        it into a Operation. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (Operation): a Operation.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return Operation(sy.serde._detail(worker, msg_tuple[1][0]), msg_tuple[1][1])


class ObjectMessage(Message):
    """Send an object to another worker using this message type.

    When a worker has an object in its local object repository (such as a tensor) and it wants
    to send that object to another worker (and delete its local copy), it uses this message type
    to do so.
    """

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.OBJ, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "ObjectMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an ObjectMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (ObjectMessage): a ObjectMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return ObjectMessage(sy.serde._detail(worker, msg_tuple[1]))


class ObjectRequestMessage(Message):
    """Request another worker to send one of its objects

    If ObjectMessage pushes an object to another worker, this Message type pulls an
    object from another worker. It also assumes that the other worker will delete it's
    local copy of the object after sending it to you."""

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.OBJ_REQ, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "ObjectRequestMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an ObjectRequestMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (ObjectRequestMessage): a ObjectRequestMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return ObjectRequestMessage(sy.serde._detail(worker, msg_tuple[1]))


class IsNoneMessage(Message):
    """Check if a worker does not have an object with a specific id.

    Occasionally we need to verify whether or not a remote worker has a specific
    object. To do so, we send an IsNoneMessage, which returns True if the object
    (such as a tensor) does NOT exist."""

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.IS_NONE, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "IsNoneMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an IsNoneMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (IsNoneMessage): a IsNoneMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return IsNoneMessage(sy.serde._detail(worker, msg_tuple[1]))


class GetShapeMessage(Message):
    """Get the shape property of a tensor in PyTorch

    We needed to have a special message type for this because .shape had some
    constraints in the older version of PyTorch."""

    # TODO: remove this message type and use ObjectRequestMessage instead.
    # note that the above to do is likely waiting for custom tensor type support in PyTorch
    # https://github.com/OpenMined/PySyft/issues/2513

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.GET_SHAPE, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "GetShapeMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an GetShapeMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (GetShapeMessage): a GetShapeMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return GetShapeMessage(sy.serde._detail(worker, msg_tuple[1]))


class ForceObjectDeleteMessage(Message):
    """Garbage collect a remote object

    This is the dominant message for garbage collection of remote objects. When
    a pointer is deleted, this message is triggered by default to tell the object
    being pointed to to also delete itself.
    """

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.FORCE_OBJ_DEL, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "ForceObjectDeleteMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an ForceObjectDeleteMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (ForceObjectDeleteMessage): a ForceObjectDeleteMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return ForceObjectDeleteMessage(sy.serde._detail(worker, msg_tuple[1]))


class SearchMessage(Message):
    """A client queries for a subset of the tensors on a remote worker using this type

    For some workers like SocketWorker we split a worker into a client and a server. For
    this configuration, a client can request to search for a subset of tensors on the server
    using this message type (this could also be called a "QueryMessage").
    """

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, contents):
        """Initialize the message using default Message constructor.

        See Message.__init__ for details."""
        super().__init__(codes.MSGTYPE.SEARCH, contents)

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "SearchMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into an SearchMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (SearchMessage): a SearchMessage.
        Examples:
            message = detail(sy.local_worker, msg_tuple)
        """
        return SearchMessage(sy.serde._detail(worker, msg_tuple[1]))


class PlanCommandMessage(Message):
    """Message used to execute a command related to plans."""

    # TODO: add more efficient detailer and simplifier custom for this type
    # https://github.com/OpenMined/PySyft/issues/2512

    def __init__(self, command_name: str, message: tuple):
        """Initialize a PlanCommandMessage.

        Args:
            command_name (str): name used to identify the command.
            message (Tuple): this is typically the args and kwargs of a method call on the client, but it
                can be any information necessary to execute the command properly.
        """

        # call the parent constructor - setting the type integer correctly
        super().__init__(codes.MSGTYPE.PLAN_CMD)

        self.command_name = command_name
        self.message = message

    @property
    def contents(self):
        """Returns a tuple with the contents of the operation (backwards compatability)."""
        return (self.command_name, self.message)

    @staticmethod
    def simplify(ptr: "PlanCommandMessage") -> tuple:
        """
        This function takes the attributes of a PlanCommandMessage and saves them in a tuple

        Args:
            ptr (PlanCommandMessage): a Message

        Returns:
            tuple: a tuple holding the unique attributes of the message
        """
        return (
            ptr.msg_type,
            (sy.serde._simplify(ptr.command_name), sy.serde._simplify(ptr.message)),
        )

    @staticmethod
    def detail(worker: AbstractWorker, msg_tuple: tuple) -> "PlanCommandMessage":
        """
        This function takes the simplified tuple version of this message and converts
        it into a PlanCommandMessage. The simplify() method runs the inverse of this method.

        Args:
            worker (AbstractWorker): a reference to the worker necessary for detailing. Read
                syft/serde/serde.py for more information on why this is necessary.
            msg_tuple (Tuple): the raw information being detailed.
        Returns:
            ptr (PlanCommandMessage): a PlanCommandMessage.
        """
        command_name, message = msg_tuple[1]
        return PlanCommandMessage(
            sy.serde._detail(worker, command_name), sy.serde._detail(worker, message)
        )



import syft as sy
import torch as th

import syft as sy
from syft.messaging import message


def test_message_serde():

    x = message.Message(0, [1, 2, 3])
    x_bin = sy.serde.serialize(x)
    y = sy.serde.deserialize(x_bin, sy.local_worker)

    assert x.contents == y.contents
    assert x.msg_type == y.msg_type


def test_cmd_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)

    y = x + x  # this is the test
    assert isinstance(bob._get_msg(-1), message.Operation)

    y = y.get()

    assert (y == th.tensor([2, 4, 6, 8])).all()

    bob.log_msgs = False


def test_obj_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)  # this is the test

    assert isinstance(bob._get_msg(-1), message.ObjectMessage)

    y = x + x

    y = y.get()

    assert (y == th.tensor([2, 4, 6, 8])).all()

    bob.log_msgs = False


def test_obj_req_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)

    y = x + x

    y = y.get()  # this is the test
    assert isinstance(bob._get_msg(-1), message.ObjectRequestMessage)

    assert (y == th.tensor([2, 4, 6, 8])).all()

    bob.log_msgs = False


def test_get_shape_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)

    y = x + x

    z = y.shape  # this is the test
    assert isinstance(bob._get_msg(-1), message.GetShapeMessage)

    assert z == th.Size([4])

    bob.log_msgs = False


def test_force_object_delete_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)

    id_on_worker = x.id_at_location

    assert id_on_worker in bob._objects

    del x  # this is the test
    assert isinstance(bob._get_msg(-1), message.ForceObjectDeleteMessage)

    assert id_on_worker not in bob._objects

    bob.log_msgs = False


def test_is_none_message(workers):

    bob = workers["bob"]

    bob.log_msgs = True

    x = th.tensor([1, 2, 3, 4]).send(bob)

    y = th.tensor([1]).send(bob)
    y.child.id_at_location = x.id_at_location

    assert not bob.request_is_remote_tensor_none(x)
    assert isinstance(bob._get_msg(-1), message.IsNoneMessage)
    assert not x.child.is_none()
    assert isinstance(bob._get_msg(-1), message.IsNoneMessage)

    del x

    assert y.child.is_none()

    bob.log_msgs = True


def test_search_message_serde():

    x = message.SearchMessage([1, 2, 3])

    x_bin = sy.serde.serialize(x)
    y = sy.serde.deserialize(x_bin, sy.local_worker)

    assert x.contents == y.contents
    assert x.msg_type == y.msg_type



import functools
import inspect
import re

import tf_encrypted as tfe


def add_constructor_registration(layer_cls):
    """
    This method rewires the layer's constructor to record arguments passed to it.
    """
    layer_cls._native_keras_constructor = layer_cls.__init__
    sig = inspect.signature(layer_cls.__init__)

    @functools.wraps(layer_cls.__init__)
    def syft_keras_constructor(self, *args, **kwargs):
        self._constructor_parameters_store = sig.bind(self, *args, **kwargs)
        self._native_keras_constructor(*args, **kwargs)

    setattr(layer_cls, "__init__", syft_keras_constructor)


def filter_layers(layers_module, tfe_layers_module):
    """
    Returns all layer types in module.
    """
    # We recognize Layer classes based on their compliance with PEP8.
    pattern = re.compile("[A-Z_][a-zA-Z0-9]+$")
    for attr_name in dir(layers_module):
        match_result = pattern.match(attr_name)
        if match_result is None:
            continue
        else:
            layer_type = match_result.group(0)
            if hasattr(tfe.keras.layers, layer_type):
                yield getattr(layers_module, layer_type)



import builtins
from multiprocessing import Process
import sys
import time

import pytest
import torch

import syft
from syft import TorchHook
from syft.generic.frameworks.hook import hook_args
from syft.workers.websocket_client import WebsocketClientWorker
from syft.workers.websocket_server import WebsocketServerWorker


def _start_proc(participant, dataset: str = None, **kwargs):  # pragma: no cover
    """Helper function for spinning up a websocket participant."""

    def target():
        server = participant(**kwargs)
        if dataset is not None:
            data, key = dataset
            server.add_dataset(data, key=key)
        server.start()

    p = Process(target=target)
    p.start()
    return p


def instantiate_websocket_client_worker(max_tries=5, sleep_time=0.1, **kwargs):  # pragma: no cover
    """Helper function to instantiate the websocket client.

    If a connection is refused, we wait a bit (`sleep_time` seconds) and try again.
    After `max_tries` failed tries, a ConnectionRefusedError is raised.
    """
    retry_counter = 0
    connection_open = False
    while not connection_open:
        try:
            remote_proxy = WebsocketClientWorker(**kwargs)
            connection_open = True
        except ConnectionRefusedError as e:
            if retry_counter < max_tries:
                retry_counter += 1
                time.sleep(sleep_time)
            else:
                raise e
    return remote_proxy


@pytest.fixture()
def start_proc():  # pragma: no cover
    return _start_proc


@pytest.fixture()
def start_remote_worker():  # pragma: no cover
    """Helper function for starting a websocket worker."""

    def _start_remote_worker(
        id, hook, dataset: str = None, host="localhost", port=8768, max_tries=5, sleep_time=0.01
    ):
        kwargs = {"id": id, "host": host, "port": port, "hook": hook}
        server = _start_proc(WebsocketServerWorker, dataset=dataset, **kwargs)
        remote_proxy = instantiate_websocket_client_worker(
            max_tries=max_tries, sleep_time=sleep_time, **kwargs
        )

        return server, remote_proxy

    return _start_remote_worker


@pytest.fixture(scope="session", autouse=True)
def hook():
    hook = TorchHook(torch)
    return hook


@pytest.fixture(scope="function", autouse=True)
def workers(hook):
    # To run a plan locally the local worker can't be a client worker,
    # since it needs to register objects
    # LaRiffle edit: doing this increases the reference count on pointers and
    # breaks the auto garbage collection for pointer of pointers, see #2150
    # hook.local_worker.is_client_worker = False

    # Reset the hook and the local worker
    syft.local_worker.clear_objects()
    hook_args.hook_method_args_functions = {}
    hook_args.hook_method_response_functions = {}
    hook_args.get_tensor_type_functions = {}

    # Define 4 virtual workers
    alice = syft.VirtualWorker(id="alice", hook=hook, is_client_worker=False)
    bob = syft.VirtualWorker(id="bob", hook=hook, is_client_worker=False)
    charlie = syft.VirtualWorker(id="charlie", hook=hook, is_client_worker=False)
    james = syft.VirtualWorker(id="james", hook=hook, is_client_worker=False)

    workers = {
        "me": hook.local_worker,
        "alice": alice,
        "bob": bob,
        "charlie": charlie,
        "james": james,
    }

    yield workers

    alice.remove_worker_from_local_worker_registry()
    bob.remove_worker_from_local_worker_registry()
    charlie.remove_worker_from_local_worker_registry()
    james.remove_worker_from_local_worker_registry()


@pytest.fixture
def hide_module():
    import_orig = builtins.__import__
    # When we check for imports in dependency_check, we don't actually attempt
    # to import each package, so popping a module from sys.modules and mocking
    # the import statement is not sufficient to simulate the dependency check
    # for when the dependency is absent. The way we check for dependencies
    # (importlib.util.find_spec) uses module Finders in the sys.meta_path when
    # checking for module specs, so we need to mock the find_spec method of the
    # Finder that will discover the module we want to hide. That Finder happens
    # to be in position three of the meta path.
    find_spec_orig = sys.meta_path[3].find_spec

    def mocked_import(name, globals, locals, fromlist, level):
        if name in ["tensorflow", "tf_encrypted", "torch"]:
            raise ImportError()

        return import_orig(name, globals, locals, fromlist, level)

    def mocked_find_spec(self, fullname, target=None):
        if self in ["tensorflow", "tf_encrypted"]:
            return None
        return find_spec_orig(self, fullname, target)

    builtins.__import__ = mocked_import
    sys.meta_path[3].find_spec = mocked_find_spec
    yield
    builtins.__import__ = import_orig
    sys.meta_path[3].find_spec = find_spec_orig



from typing import Union
import weakref

import torch

import syft as sy
from syft.generic.pointers.object_wrapper import ObjectWrapper
from syft.workers.abstract import AbstractWorker
from syft.workers.base import BaseWorker


class TrainConfig:
    """TrainConfig abstraction.

    A wrapper object that contains all that is needed to run a training loop
    remotely on a federated learning setup.
    """

    def __init__(
        self,
        model: torch.jit.ScriptModule,
        loss_fn: torch.jit.ScriptModule,
        owner: AbstractWorker = None,
        batch_size: int = 32,
        epochs: int = 1,
        optimizer: str = "SGD",
        optimizer_args: dict = {"lr": 0.1},
        id: Union[int, str] = None,
        max_nr_batches: int = -1,
        shuffle: bool = True,
        loss_fn_id: int = None,
        model_id: int = None,
    ):
        """Initializer for TrainConfig.

        Args:
            model: A traced torch nn.Module instance.
            loss_fn: A jit function representing a loss function which
                shall be used to calculate the loss.
            batch_size: Batch size used for training.
            epochs: Epochs used for training.
            optimizer: A string indicating which optimizer should be used.
            optimizer_args: A dict containing the arguments to initialize the optimizer. Defaults to {'lr': 0.1}.
            owner: An optional BaseWorker object to specify the worker on which
                the tensor is located.
            id: An optional string or integer id of the tensor.
            max_nr_batches: Maximum number of training steps that will be performed. For large datasets
                            this can be used to run for less than the number of epochs provided.
            shuffle: boolean, whether to access the dataset randomly (shuffle) or sequentially (no shuffle).
            loss_fn_id: The id_at_location of (the ObjectWrapper of) a loss function which
                        shall be used to calculate the loss. This is used internally for train config deserialization.
            model_id: id_at_location of a traced torch nn.Module instance (objectwrapper). . This is used internally for train config deserialization.
        """
        # syft related attributes
        self.owner = owner if owner else sy.hook.local_worker
        self.id = id if id is not None else sy.ID_PROVIDER.pop()
        self.location = None

        # training related attributes
        self.model = model
        self.loss_fn = loss_fn
        self.batch_size = batch_size
        self.epochs = epochs
        self.optimizer = optimizer
        self.optimizer_args = optimizer_args
        self.max_nr_batches = max_nr_batches
        self.shuffle = shuffle

        # pointers
        self.model_ptr = None
        self.loss_fn_ptr = None

        # internal ids
        self._model_id = model_id
        self._loss_fn_id = loss_fn_id

    def __str__(self) -> str:
        """Returns the string representation of a TrainConfig."""
        out = "<"
        out += str(type(self)).split("'")[1].split(".")[-1]
        out += " id:" + str(self.id)
        out += " owner:" + str(self.owner.id)

        if self.location:
            out += " location:" + str(self.location.id)

        out += " epochs: " + str(self.epochs)
        out += " batch_size: " + str(self.batch_size)
        out += " optimizer_args: " + str(self.optimizer_args)

        out += ">"
        return out

    def _wrap_and_send_obj(self, obj, location):
        """Wrappers object and send it to location."""
        obj_with_id = ObjectWrapper(id=sy.ID_PROVIDER.pop(), obj=obj)
        obj_ptr = self.owner.send(obj_with_id, location)
        obj_id = obj_ptr.id_at_location
        return obj_ptr, obj_id

    def send(self, location: BaseWorker) -> weakref:
        """Gets the pointer to a new remote object.

        One of the most commonly used methods in PySyft, this method serializes
        the object upon which it is called (self), sends the object to a remote
        worker, creates a pointer to that worker, and then returns that pointer
        from this function.

        Args:
            location: The BaseWorker object which you want to send this object
                to. Note that this is never actually the BaseWorker but instead
                a class which instantiates the BaseWorker abstraction.
        Returns:
            A weakref instance.
        """
        # Send traced model
        self.model_ptr, self._model_id = self._wrap_and_send_obj(self.model, location)

        # Send loss function
        self.loss_fn_ptr, self._loss_fn_id = self._wrap_and_send_obj(self.loss_fn, location)

        # Send train configuration itself
        ptr = self.owner.send(self, location)

        return ptr

    def get(self, location):
        return self.owner.request_obj(self, location)

    def get_model(self):
        if self.model is not None:
            return self.model_ptr.get()

    def get_loss_fn(self):
        if self.loss_fn is not None:
            return self.loss_fn.get()

    @staticmethod
    def simplify(train_config: "TrainConfig") -> tuple:
        """Takes the attributes of a TrainConfig and saves them in a tuple.

        Attention: this function does not serialize the model and loss_fn attributes
        of a TrainConfig instance, these are serialized and sent before. TrainConfig
        keeps a reference to the sent objects using _model_id and _loss_fn_id which
        are serialized here.

        Args:
            train_config: a TrainConfig object
        Returns:
            tuple: a tuple holding the unique attributes of the TrainConfig object
        """
        return (
            train_config._model_id,
            train_config._loss_fn_id,
            train_config.batch_size,
            train_config.epochs,
            sy.serde._simplify(train_config.optimizer),
            sy.serde._simplify(train_config.optimizer_args),
            sy.serde._simplify(train_config.id),
            train_config.max_nr_batches,
            train_config.shuffle,
        )

    @staticmethod
    def detail(worker: AbstractWorker, train_config_tuple: tuple) -> "TrainConfig":
        """This function reconstructs a TrainConfig object given it's attributes in the form of a tuple.

        Args:
            worker: the worker doing the deserialization
            train_config_tuple: a tuple holding the attributes of the TrainConfig
        Returns:
            train_config: A TrainConfig object
        """

        model_id, loss_fn_id, batch_size, epochs, optimizer, optimizer_args, id, max_nr_batches, shuffle = (
            train_config_tuple
        )

        id = sy.serde._detail(worker, id)
        detailed_optimizer = sy.serde._detail(worker, optimizer)
        detailed_optimizer_args = sy.serde._detail(worker, optimizer_args)

        train_config = TrainConfig(
            model=None,
            loss_fn=None,
            owner=worker,
            id=id,
            model_id=model_id,
            loss_fn_id=loss_fn_id,
            batch_size=batch_size,
            epochs=epochs,
            optimizer=detailed_optimizer,
            optimizer_args=detailed_optimizer_args,
            max_nr_batches=max_nr_batches,
            shuffle=shuffle,
        )

        return train_config



import syft

if syft.dependency_check.tensorflow_available:
    from syft_tensorflow.hook import TensorFlowHook
    from syft_tensorflow.tensor import TensorFlowTensor

    setattr(syft, "TensorFlowHook", TensorFlowHook)



import pytest
import torch
import syft as sy
from test.efficiency_tests.assertions import assert_time


@pytest.mark.parametrize("activation", ["tanh", "sigmoid"])
@assert_time(max_time=1)
def test_activation(activation, hook, workers):

    activation_func = torch.tanh if activation == "tanh" else torch.sigmoid

    bob = workers["bob"]
    alice = workers["alice"]
    crypto_prov = sy.VirtualWorker(hook, id="crypto_prov")

    x = torch.randn([10, 10]).fix_precision().share(bob, alice, crypto_provider=crypto_prov)
    activation_func(x)



import unittest.mock as mock
import pytest

from syft import exceptions
from syft.generic import id_provider


def test_pop_no_given_ids(hook):
    provider = id_provider.IdProvider()
    values = [10, 4, 15, 4, 2, 0]

    orig_func = id_provider.create_random_id
    mocked_random_numbers = mock.Mock()
    mocked_random_numbers.side_effect = values
    id_provider.create_random_id = mocked_random_numbers

    val = provider.pop()
    assert val == values[0]

    val = provider.pop()
    assert val == values[1]

    val = provider.pop()
    assert val == values[2]

    # values[3] is skipped, as value already used.

    val = provider.pop()
    assert val == values[4]

    val = provider.pop()
    assert val == values[5]

    id_provider.create_random_id = orig_func


def test_pop_with_given_ids(hook):
    given_ids = [4, 15, 2]
    provider = id_provider.IdProvider(given_ids=given_ids.copy())
    values = [10, 4, 15, 4, 2, 0]

    orig_func = id_provider.create_random_id
    mocked_random_numbers = mock.Mock()
    mocked_random_numbers.side_effect = values
    id_provider.create_random_id = mocked_random_numbers

    val = provider.pop()
    assert val == given_ids[-1]

    val = provider.pop()
    assert val == given_ids[-2]

    val = provider.pop()
    assert val == given_ids[-3]

    val = provider.pop()
    assert val == values[0]

    # values[1, 2, 3, 4] are skipped, as value already used.

    val = provider.pop()
    assert val == values[5]

    id_provider.create_random_id = orig_func


def test_given_ids_side_effect(hook):
    given_ids = [4, 15, 2]
    provider = id_provider.IdProvider(given_ids=given_ids)

    assert len(given_ids) == 3
    provider.pop()

    assert len(given_ids) == 2

    provider.pop()
    assert len(given_ids) == 1

    provider.pop()
    assert len(given_ids) == 0


def test_set_next_ids(hook):
    initial_given_ids = [2, 3]
    provider = id_provider.IdProvider(given_ids=initial_given_ids.copy())

    next_ids = [4, 5]
    provider.set_next_ids(next_ids.copy())

    val = provider.pop()
    assert val == next_ids[-1]
    val = provider.pop()
    assert val == next_ids[-2]

    val = provider.pop()
    assert val == initial_given_ids[-1]
    val = provider.pop()
    assert val == initial_given_ids[-2]


def test_set_next_ids_with_id_checking(hook):
    initial_given_ids = [2, 3]
    provider = id_provider.IdProvider()
    provider.set_next_ids(initial_given_ids.copy(), check_ids=False)

    # generated the initial 3 ids
    provider.pop()
    provider.pop()
    provider.pop()

    next_ids = [1, 2, 5]
    with pytest.raises(exceptions.IdNotUniqueError, match=r"\{2\}"):
        provider.set_next_ids(next_ids.copy(), check_ids=True)

    next_ids = [2, 3, 5]
    with pytest.raises(exceptions.IdNotUniqueError, match=r"\{2, 3\}"):
        provider.set_next_ids(next_ids.copy(), check_ids=True)


def test_start_recording_ids():
    initial_given_ids = [2, 3]
    provider = id_provider.IdProvider(given_ids=initial_given_ids.copy())
    provider.pop()
    provider.start_recording_ids()
    provider.pop()

    ids = provider.get_recorded_ids()
    assert len(ids) == 1
    assert ids[0] == initial_given_ids[-2]


def test_get_recorded_ids():
    initial_given_ids = [2, 3, 4]
    provider = id_provider.IdProvider(given_ids=initial_given_ids.copy())
    provider.pop()
    provider.start_recording_ids()
    provider.pop()

    ids = provider.get_recorded_ids(continue_recording=True)
    assert len(ids) == 1
    assert ids[0] == initial_given_ids[-2]

    provider.pop()

    ids = provider.get_recorded_ids()
    assert len(ids) == 2
    assert ids[0] == initial_given_ids[-2]
    assert ids[1] == initial_given_ids[-3]



import pytest
import torch
from torch import Tensor
import syft as sy


def test_virtual_grid(workers):
    """This tests our ability to simplify tuple types.

    This test is pretty simple since tuples just serialize to
    themselves, with a tuple wrapper with the correct ID (1)
    for tuples so that the detailer knows how to interpret it."""

    print(len(workers))
    print(workers)

    bob = workers["bob"]
    alice = workers["alice"]
    james = workers["james"]

    grid = sy.grid.VirtualGrid(*[bob, alice, james])

    x = torch.tensor([1, 2, 3, 4]).tag("#bob", "#male").send(bob)
    y = torch.tensor([1, 2, 3, 4]).tag("#alice", "#female").send(alice)
    z = torch.tensor([1, 2, 3, 4]).tag("#james", "#male").send(james)

    results, tags = grid.search()
    assert len(results) == 3

    assert "bob" in results.keys()
    assert "alice" in results.keys()
    assert "james" in results.keys()

    results, tags = grid.search("#bob")
    assert len(results["bob"]) == 1
    assert "alice" not in results
    assert "james" not in results

    results, tags = grid.search("#male")
    assert len(results["bob"]) == 1
    assert "alice" not in results
    assert len(results["james"]) == 1



from functools import wraps
import torch

import syft
from syft.generic.tensor import AbstractTensor
from syft.generic.frameworks.hook import hook_args
from syft.generic.frameworks.overload import overloaded
from . import gradients


def backwards_grad(grad_fn, in_grad=None):
    if grad_fn is None:
        raise ValueError(
            "The gradient for one of the command you used was not found. Check gradients.py "
            "to see if it's missing."
        )
    back_grad = grad_fn(in_grad)
    for next_grad_fn, next_grad in zip(grad_fn.next_functions, back_grad):
        backwards_grad(next_grad_fn, next_grad)


class AutogradTensor(AbstractTensor):
    """ A tensor that tracks operations to build a dynamic graph and backprops
        through the graph to calculate gradients.
    """

    def __init__(
        self, data=None, requires_grad=True, owner=None, id=None, preinitialize_grad=False, **kwargs
    ):
        super().__init__()

        self.owner = owner
        self.id = id

        self.child = data
        self.requires_grad = requires_grad
        self.preinitialize_grad = preinitialize_grad

        if preinitialize_grad:
            self.grad = data * 0
        else:
            self.grad = None
        self.grad_fn = None

    def backward(self, grad=None):
        if grad is None:
            # Build a torch tensor of ones with the same shape
            # And chain structure than self
            grad = self * 0 + 1
        backwards_grad(self.grad_fn, grad)

    @property
    def data(self):
        return self

    @data.setter
    def data(self, new_data):
        self.child = new_data.child
        return self

    @property
    def grad(self):
        return self._grad

    @grad.setter
    def grad(self, value):
        self._grad = value

    def attr(self, attr_name):
        if attr_name == "grad":
            return self.grad

        attr_val = self.child.attr(attr_name)
        return attr_val

    def __add__(self, other):
        return self.add(other)

    def __iadd__(self, other):
        result = self.add(other)
        self.child = result.child
        self.grad_fn = result.grad_fn

    def __sub__(self, other):
        return self.sub(other)

    def __isub__(self, other):
        result = self.sub(other)
        self.child = result.child
        self.grad_fn = result.grad_fn

    def __mul__(self, other):
        return self.mul(other)

    def __matmul__(self, other):
        return self.matmul(other)

    def __pow__(self, power, **kwargs):
        return self.pow(power, **kwargs)

    def __truediv__(self, other):
        return self.div(other)

    @overloaded.method
    def __gt__(self, _self, other):
        return _self.__gt__(other)

    @overloaded.method
    def __ge__(self, _self, other):
        return _self.__ge__(other)

    @overloaded.method
    def __lt__(self, _self, other):
        return _self.__lt__(other)

    @overloaded.method
    def __le__(self, _self, other):
        return _self.__le__(other)

    @overloaded.method
    def eq(self, _self, other):
        return _self.eq(other)

    @overloaded.method
    def relu(self, self_):
        return self_.relu()

    def __getattribute__(self, name):
        # Automatically attaching gradient functions if they are defined in the
        # gradients module.
        grad_fn = getattr(gradients, name.capitalize() + "Backward", None)

        # print(f"getattribute {name}")
        if grad_fn is not None:

            def method_with_grad(*args, **kwargs):
                new_self, new_args, new_kwargs = hook_args.unwrap_args_from_method(
                    name, self, args, kwargs
                )

                result = getattr(new_self, name)(*new_args, **new_kwargs)

                # Put back SyftTensor on the tensors found in the response
                result = hook_args.hook_response(name, result, wrap_type=type(self))
                result.grad_fn = grad_fn(self, *args, **kwargs)
                result.grad_fn.result = result

                return result

            return method_with_grad
        else:
            return object.__getattribute__(self, name)

    @staticmethod
    @overloaded.module
    def torch(module):
        def add(self, other):
            return self.add(other)

        module.add = add

        def sub(self, other):
            return self.sub(other)

        module.sub = sub

        def mul(self, other):
            return self.mul(other)

        module.mul = mul

        def matmul(self, other):
            return self.matmul(other)

        module.matmul = matmul

        def div(self, other):
            return self.div(other)

        module.div = div

        def addmm(bias, input_tensor, weight):
            matmul = input_tensor.matmul(weight)
            result = bias.add(matmul)
            return result

        module.addmm = addmm

        @overloaded.module
        def nn(module):
            """
            The syntax is the same, so @overloaded.module handles recursion
            Note that we don't need to add the @staticmethod decorator
            """

            @overloaded.module
            def functional(module):
                def linear(*args):
                    """
                    Un-hook the function to have its detailed behaviour
                    """
                    return torch.nn.functional.native_linear(*args)

                module.linear = linear

                def relu(tensor):
                    return tensor.relu()

                module.relu = relu

            module.functional = functional

        # Modules should be registered just like functions
        module.nn = nn

    @classmethod
    def handle_func_command(cls, command):
        """
        Receive an instruction for a function to be applied on a AutogradTensor,
        Perform some specific action (like logging) which depends of the
        instruction content, replace in the args all the LogTensors with
        their child attribute, forward the command instruction to the
        handle_function_command of the type of the child attributes, get the
        response and replace a AutogradTensor on top of all tensors found in
        the response.
        :param command: instruction of a function command: (command name,
        <no self>, arguments[, kwargs])
        :return: the response of the function command
        """

        cmd, _, args, kwargs = command

        # Check that the function has not been overwritten
        try:
            # Try to get recursively the attributes in cmd = "<attr1>.<attr2>.<attr3>..."
            cmd = cls.rgetattr(cls, cmd)
            return cmd(*args, **kwargs)
        except AttributeError:
            pass

        # Replace all AutogradTensor with their child attribute
        new_args, new_kwargs, new_type = hook_args.unwrap_args_from_function(cmd, args, kwargs)

        # build the new command
        new_command = (cmd, None, new_args, new_kwargs)

        # Send it to the appropriate class and get the response
        response = new_type.handle_func_command(new_command)

        # Put back AutogradTensor on the tensors found in the response
        response = hook_args.hook_response(cmd, response, wrap_type=cls)

        return response

    def get(self):
        """Just a pass through. This is most commonly used when calling .get() on a
        AutogradTensor which has also been shared."""
        tensor = self.child.get()

        if isinstance(tensor, torch.Tensor):
            # Remove the autograd node if a simple tensor is received
            if not tensor.is_wrapper:
                return tensor
            # If it's a wrapper, then insert the autograd under the wrapper
            else:
                self.child = tensor.child
                tensor.child = self
                return tensor

        self.child = tensor
        return self

    def float_precision(self):
        """Just a pass through. This is most commonly used when calling .float_precision() on a
        AutogradTensor which has also been shared."""
        self.child = self.child.float_precision()
        # Remove the autograd node if a simple tensor is received
        if isinstance(self.child, torch.Tensor) and not self.child.is_wrapper:
            return self.child
        return self

    @staticmethod
    def simplify(tensor: "AutogradTensor") -> tuple:
        """Takes the attributes of an AutogradTensor and saves them in a tuple.
            Or simply said, it serializes an AutogradTensor
        Args:
            tensor: an AutogradTensor.

        Returns:
            tuple: a tuple holding the unique attributes of the AutogradTensor.
        """
        chain = syft.serde._simplify(tensor.child) if hasattr(tensor, "child") else None

        return (
            tensor.owner,
            syft.serde._simplify(tensor.id),
            chain,
            tensor.requires_grad,
            tensor.preinitialize_grad,
            tensor.grad_fn,
            # tensor.local_autograd,
            syft.serde._simplify(tensor.tags),
            syft.serde._simplify(tensor.description),
        )

    @staticmethod
    def detail(worker: AbstractTensor, tensor_tuple: tuple) -> "AutogradTensor":
        """
            This function reconstructs (deserializes) an AutogradTensors given its attributes in form of a tuple.
            Args:
                worker: the worker doing the deserialization
                tensor_tuple: a tuple holding the attributes of the AutogradTensor
            Returns:
                AutogradTensor: an AutogradTensor
            Examples:
                shared_tensor = detail(data)
            """
        owner, tensor_id, chain, requires_grad, preinitialize_grad, grad_fn, tags, description = (
            tensor_tuple
        )

        if chain is not None:
            chain = syft.serde._detail(worker, chain)

        tensor = AutogradTensor(
            owner=owner,
            id=syft.serde._detail(worker, tensor_id),
            requires_grad=requires_grad,  # ADDED!
            preinitialize_grad=preinitialize_grad,
            grad_fn=grad_fn,
            # local_autograd=local_autograd,
            data=chain,  # pass the de-serialized data
            tags=syft.serde._detail(worker, tags),
            description=syft.serde._detail(worker, description),
        )

        return tensor



from typing import List
from typing import Union
from typing import TYPE_CHECKING

import syft
from syft import exceptions
from syft.messaging.message import ForceObjectDeleteMessage
from syft.generic.frameworks.hook import hook_args
from syft.generic.object import AbstractObject

# this if statement avoids circular imports between base.py and pointer.py
if TYPE_CHECKING:
    from syft.workers.base import BaseWorker


class ObjectPointer(AbstractObject):
    """A pointer to a remote object.

    An ObjectPointer forwards all API calls to the remote. ObjectPointer objects
    point to objects. They exist to mimic the entire
    API of an object, but instead of computing a function locally
    (such as addition, subtraction, etc.) they forward the computation to a
    remote machine as specified by self.location. Specifically, every
    ObjectPointer has a object located somewhere that it points to (they should
    never exist by themselves).
    The objects being pointed to can be on the same machine or (more commonly)
    on a different one. Note further that a ObjectPointer does not know the
    nature how it sends messages to the object it points to (whether over
    socket, http, or some other protocol) as that functionality is abstracted
    in the BaseWorker object in self.location.
    """

    def __init__(
        self,
        location: "BaseWorker" = None,
        id_at_location: Union[str, int] = None,
        owner: "BaseWorker" = None,
        id: Union[str, int] = None,
        garbage_collect_data: bool = True,
        point_to_attr: str = None,
        tags: List[str] = None,
        description: str = None,
    ):

        """Initializes a ObjectPointer.

        Args:
            location: An optional BaseWorker object which points to the worker
                on which this pointer's object can be found.
            id_at_location: An optional string or integer id of the object
                being pointed to.
            owner: An optional BaseWorker object to specify the worker on which
                the pointer is located. It is also where the pointer is
                registered if register is set to True. Note that this is
                different from the location parameter that specifies where the
                pointer points to.
            id: An optional string or integer id of the PointerTensor.
            garbage_collect_data: If true (default), delete the remote object when the
                pointer is deleted.
            point_to_attr: string which can tell a pointer to not point directly to\
                an object, but to point to an attribute of that object such as .child or
                .grad. Note the string can be a chain (i.e., .child.child.child or
                .grad.child.child). Defaults to None, which means don't point to any attr,
                just point to then object corresponding to the id_at_location.
        """
        super().__init__(id=id, owner=owner, tags=tags, description=description)

        self.location = location
        self.id_at_location = id_at_location
        self.garbage_collect_data = garbage_collect_data
        self.point_to_attr = point_to_attr

    @classmethod
    def handle_func_command(cls, command):
        """
        Receive an instruction for a function to be applied on a Pointer,
        Get the remote location to send the command, send it and get a
        pointer to the response, return.
        :param command: instruction of a function command: (command name,
        None, arguments[, kwargs])
        :return: the response of the function command
        """
        pointer = cls.find_a_pointer(command)
        # Get info on who needs to send where the command
        owner = pointer.owner
        location = pointer.location

        # Send the command
        response = owner.send_command(location, command)

        return response

    @classmethod
    def find_a_pointer(cls, command):
        """
        Find and return the first pointer in the args object, using a trick
        with the raising error RemoteObjectFoundError
        """
        try:
            cmd, _, args, kwargs = command
            _ = hook_args.unwrap_args_from_function(cmd, args, kwargs)
        except exceptions.RemoteObjectFoundError as err:
            pointer = err.pointer
            return pointer

    def get(self, deregister_ptr: bool = True):
        """Requests the object being pointed to.

        The object to which the pointer points will be requested, serialized and returned.

        Note:
            This will typically mean that the remote object will be
            removed/destroyed.

        Args:
            deregister_ptr (bool, optional): this determines whether to
                deregister this pointer from the pointer's owner during this
                method. This defaults to True because the main reason people use
                this method is to move the tensor from the location to the
                local one, at which time the pointer has no use.

        Returns:
            An AbstractObject object which is the tensor (or chain) that this
            object used to point to on a location.

        TODO: add param get_copy which doesn't destroy remote if true.
        """

        if self.point_to_attr is not None:

            raise exceptions.CannotRequestObjectAttribute(
                "You called .get() on a pointer to"
                " a tensor attribute. This is not yet"
                " supported. Call .clone().get() instead."
            )

        # if the pointer happens to be pointing to a local object,
        # just return that object (this is an edge case)
        if self.location == self.owner:
            obj = self.owner.get_obj(self.id_at_location)
            if hasattr(obj, "child"):
                obj = obj.child
        else:
            # get tensor from location
            obj = self.owner.request_obj(self.id_at_location, self.location)

        # Remove this pointer by default
        if deregister_ptr:
            self.owner.de_register_obj(self)

        if self.garbage_collect_data:
            # data already retrieved, do not collect any more.
            self.garbage_collect_data = False

        return obj

    def __str__(self):
        """Returns a string version of this pointer.

        This is primarily for end users to quickly see things about the object.
        This tostring shouldn't be used for anything else though as it's likely
        to change. (aka, don't try to parse it to extract information. Read the
        attribute you need directly). Also, don't use this to-string as a
        serialized form of the pointer.
        """

        type_name = type(self).__name__
        out = (
            f"["
            f"{type_name} | "
            f"{str(self.owner.id)}:{self.id}"
            " -> "
            f"{str(self.location.id)}:{self.id_at_location}"
            f"]"
        )

        if self.point_to_attr is not None:
            out += "::" + str(self.point_to_attr).replace(".", "::")

        big_str = False

        if self.tags is not None and len(self.tags):
            big_str = True
            out += "\n\tTags: "
            for tag in self.tags:
                out += str(tag) + " "

        if big_str and hasattr(self, "shape"):
            out += "\n\tShape: " + str(self.shape)

        if self.description is not None:
            big_str = True
            out += "\n\tDescription: " + str(self.description).split("\n")[0] + "..."

        return out

    def __repr__(self):
        """Returns the to-string method.

        When called using __repr__, most commonly seen when returned as cells
        in Jupyter notebooks.
        """
        return self.__str__()

    def __del__(self):
        """This method garbage collects the object this pointer is pointing to.
        By default, PySyft assumes that every object only has one pointer to it.
        Thus, if the pointer gets garbage collected, we want to automatically
        garbage collect the object being pointed to.
        """

        # if .get() gets called on the pointer before this method is called, then
        # the remote object has already been removed. This results in an error on
        # this next line because self no longer has .owner. Thus, we need to check
        # first here and not try to call self.owner.anything if self doesn't have
        # .owner anymore.
        if hasattr(self, "owner") and self.garbage_collect_data:
            # attribute pointers are not in charge of GC
            if self.point_to_attr is None:
                self.owner.send_msg(ForceObjectDeleteMessage(self.id_at_location), self.location)

    def _create_attr_name_string(self, attr_name):
        if self.point_to_attr is not None:
            point_to_attr = "{}.{}".format(self.point_to_attr, attr_name)
        else:
            point_to_attr = attr_name
        return point_to_attr

    def attr(self, attr_name):
        attr_ptr = syft.ObjectPointer(
            id=self.id,
            owner=self.owner,
            location=self.location,
            id_at_location=self.id_at_location,
            point_to_attr=self._create_attr_name_string(attr_name),
        )  # .wrap()
        self.__setattr__(attr_name, attr_ptr)
        return attr_ptr

    def setattr(self, name, value):
        self.owner.send_command(
            message=("__setattr__", self, (name, value), {}), recipient=self.location
        )
